from __future__ import annotations
import random
import time
import warnings
from typing import Optional, Union

from . import aws_connection, connection, mqtt_connection
from .model import (
    JciHitachiAC,
    JciHitachiACSupport,
    JciHitachiAWSStatus,
    JciHitachiAWSStatusSupport,
    JciHitachiDH,
    JciHitachiDHSupport,
    JciHitachiStatus,
    JciHitachiStatusSupport,
)
from .status import (
    JciHitachiCommand,
    JciHitachiCommandAC,
    JciHitachiCommandDH,
    JciHitachiStatusInterpreter,
)


class Peripheral:  # pragma: no cover
    """Peripheral (Device) Information.

    Parameters
    ----------
    peripheral_json : dict
        Peripheral json of specific device.
    """

    supported_device_type = {
        144: "AC",
        146: "DH",
        # 148: "HE",
    }

    def __init__(self, peripheral_json: dict) -> None:
        self._json = peripheral_json
        self._available = False
        self._status_code = ""
        self._support_code = ""
        self._supported_status = None

        warnings.warn("This API has been deprecated.", DeprecationWarning)

    def __repr__(self) -> str:
        ret = (
            f"name: {self.name}\n"
            f"brand: {self.brand}\n"
            f"model: {self.model}\n"
            f"type: {self.type}\n"
            f"available: {self.available}\n"
            f"status_code: {self.status_code}\n"
            f"support_code: {self.support_code}\n"
            f"gateway_id: {self.gateway_id}\n"
            f"gateway_mac_address: {self.gateway_mac_address}"
        )
        return ret

    @classmethod
    def from_device_names(
        cls, peripherals_json: dict, device_names: Optional[Union[list[str], str]]
    ) -> dict[str, object]:
        """Use device names to pick peripheral_jsons accordingly.

        Parameters
        ----------
        peripherals_json : dict
            Peripherals_json retrieved from GetPeripheralsByUser.
        device_names : list of str or str or None
            Device name. If None is given, all available devices will be included, by default None.

        Returns
        -------
        dict
            A dict of Peripheral instances with device name key.
        """

        peripherals = {}

        if isinstance(device_names, str):
            device_names = [device_names]

        for result in peripherals_json["results"]:
            device_name = result["DeviceName"]
            device_type = result["Peripherals"][0]["DeviceType"]

            if device_names is None or (device_names and device_name in device_names):
                if device_type in cls.supported_device_type:
                    peripherals[device_name] = cls(result)

        assert device_names is None or len(device_names) == len(
            peripherals
        ), "Some of device_names are not available from the API."

        return peripherals

    @property
    def available(self) -> bool:
        """Whether the device is available.

        Returns
        -------
        bool
            Return True if the device is available.
        """

        return self._available

    @available.setter
    def available(self, x) -> None:
        self._available = x

    @property
    def brand(self) -> str:
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """

        return getattr(self.supported_status, "brand")

    @property
    def commander(self) -> Union[JciHitachiCommand, None]:
        """Return a new JciHitachiCommand instance based on peripheral's type.

        Returns
        -------
        JciHitachiCommand or None
            JciHitachiCommand instance.
        """

        if self.type == "AC":
            return JciHitachiCommandAC(self.gateway_mac_address)
        elif self.type == "DH":
            return JciHitachiCommandDH(self.gateway_mac_address)
        else:
            return None

    @property
    def gateway_id(self) -> int:
        """Gateway ID.

        Returns
        -------
        int
            Gateway ID.
        """

        return self._json["ObjectID"]

    @property
    def gateway_mac_address(self) -> str:
        """Gateway mac address.

        Returns
        -------
        str
            Gateway mac address.
        """

        return self._json["GMACAddress"]

    @property
    def model(self) -> str:
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        return getattr(self.supported_status, "model")

    @property
    def name(self) -> str:
        """Device name.

        Returns
        -------
        str
            Device name.
        """

        return self._json["DeviceName"]

    @property
    def picked_peripheral(self) -> dict:
        """Picked peripheral.

        Returns
        -------
        dict
            Picked peripheral.
        """

        return self._json

    @property
    def status_code(self) -> str:
        """Peripheral's status code (LValue) reported by the API.

        Returns
        -------
        str
            Status code.
        """

        return self._status_code

    @status_code.setter
    def status_code(self, x: str) -> None:
        self._status_code = x

    @property
    def support_code(self) -> str:
        """Peripheral's support code (LValue) reported by the API.

        Returns
        -------
        str
            Status code.
        """

        return self._support_code

    @support_code.setter
    def support_code(self, x: str) -> None:
        self._support_code = x
        if self.type == "AC":
            self._supported_status = JciHitachiACSupport(
                JciHitachiStatusInterpreter(x).decode_support()
            )
        elif self.type == "DH":
            self._supported_status = JciHitachiDHSupport(
                JciHitachiStatusInterpreter(x).decode_support()
            )

    @property
    def supported_status(self) -> JciHitachiStatusSupport:
        """Peripheral's supported status converted from support_code.

        Returns
        -------
        JciHitachiStatusSupport
            Supported status.
        """

        return self._supported_status

    @property
    def type(self) -> str:
        """Device type.

        Returns
        -------
        str
            Device type.
            If not supported, 'unknown' will be returned. (currently supports: `AC`, `DH`)
        """

        return self.supported_device_type.get(
            self._json["Peripherals"][0]["DeviceType"], "unknown"
        )


class JciHitachiAPI:  # pragma: no cover
    """Jci-Hitachi API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    device_names : list of str or str or None, optional
        Device names. If None is given, all available devices will be included, by default None.
    max_retries : int, optional
        Maximum number of retries when setting status, by default 5.
    device_offline_timeout: float, optional
        Device offline timeout, by default 45.0.
    print_response : bool, optional
        If set, all responses of httpx and MQTT will be printed, by default False.
    """

    def __init__(
        self,
        email: str,
        password: str,
        device_names: Optional[Union[list[str], str]] = None,
        max_retries: int = 5,
        device_offline_timeout: float = 45.0,
        print_response: bool = False,
    ) -> None:
        self.email = email
        self.password = password
        self.device_names = device_names
        self.max_retries = max_retries
        self.device_offline_timeout = device_offline_timeout
        self.print_response = print_response

        self._mqtt = None
        self._device_id = random.randint(1000, 6999)
        self._peripherals = {}
        self._session_token = None
        self._user_id = None
        self._task_id = 0

        warnings.warn("This API has been deprecated.", DeprecationWarning)

    @property
    def peripherals(self) -> dict[str, Peripheral]:
        """Picked peripherals.

        Returns
        -------
        dict
            A dict of Peripherals.
        """

        return self._peripherals

    @property
    def user_id(self) -> Optional[int]:
        """User ID.

        Returns
        -------
        int
            User ID.
        """

        return self._user_id

    @property
    def task_id(self) -> int:
        """Task ID.

        Returns
        -------
        int
            Serial number counted from 0, with maximum 999.
        """

        self._task_id += 1
        if self._task_id >= 1000:
            self._task_id = 1

        return self._task_id

    def _sync_peripherals_availablity(self) -> None:
        device_access_time = self._mqtt.mqtt_events.device_access_time
        for peripheral in self._peripherals.values():
            if (
                peripheral.gateway_id in device_access_time
                and abs(time.time() - device_access_time[peripheral.gateway_id])
                < self.device_offline_timeout
            ):
                peripheral.available = True
            else:
                peripheral.available = False

    def login(self) -> None:
        """Login API.

        Raises
        ------
        RuntimeError
            If a login error occurs, RuntimeError will be raised.
        """

        conn = connection.GetPeripheralsByUser(
            self.email, self.password, print_response=self.print_response
        )
        conn_status, conn_json = conn.get_data()
        self._session_token = conn.session_token

        if conn_status == "OK":
            if len(conn_json["results"]) != 0:
                self._user_id = conn_json["results"][0]["Owner"]

            # peripherals
            self._peripherals = Peripheral.from_device_names(
                conn_json, self.device_names
            )
            self.device_names = list(self._peripherals.keys())

            # mqtt
            self._mqtt = mqtt_connection.JciHitachiMqttConnection(
                self.email,
                self.password,
                self._user_id,
                print_response=self.print_response,
            )
            self._mqtt.configure()
            self._mqtt.connect()

            # status
            self.refresh_status()

        else:
            raise RuntimeError(f"An error occurred when API login: {conn_status}.")

    def logout(self) -> None:
        """Logout API."""

        self._mqtt.disconnect()

    def change_password(self, new_password: str) -> None:
        """Change password.

        Parameters
        ----------
        new_password : str
            New password.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        conn = connection.UpdateUserCredential(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response,
        )
        conn_status, conn_json = conn.get_data(new_password)
        self._session_token = conn.session_token

    def get_status(
        self, device_name: Optional[str] = None
    ) -> dict[str, JciHitachiStatus]:
        """Get device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.

        Returns
        -------
        dict of JciHitachiStatus.
            Return a dict of JciHitachiStatus instances according to device type.
            For example, if the device type is `AC`, then return JciHitachiAC instance.
        """

        statuses = {}
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or peripheral.type == "unknown":
                continue
            dev_status = JciHitachiStatusInterpreter(
                peripheral.status_code
            ).decode_status()

            if peripheral.type == "AC":
                statuses[name] = JciHitachiAC(dev_status)
            elif peripheral.type == "DH":
                statuses[name] = JciHitachiDH(dev_status)
        return statuses

    def get_supported_status(
        self, device_name: Optional[str] = None
    ) -> dict[str, JciHitachiStatusSupport]:
        """Get supported device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.

        Returns
        -------
        dict of JciHitachiStatusSupport.
            Return a dict of JciHitachiStatusSupport instances according to device type.
            For example, if the device type is `AC`, then return JciHitachiACSupport instance.
        """

        supported_statuses = {}
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or peripheral.type == "unknown":
                continue
            supported_statuses[name] = peripheral.supported_status

        return supported_statuses

    def refresh_status(self, device_name: Optional[str] = None) -> None:
        """Refresh device status from the API.

        Parameters
        ----------
        device_name : str, optional
            Refreshing a device's status by its name.
            If None is given, all devices' status will be refreshed,
            by default None.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._sync_peripherals_availablity()

        conn = connection.GetDataContainerByID(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response,
        )
        for name, peripheral in self._peripherals.items():
            if (device_name and name != device_name) or peripheral.type == "unknown":
                continue
            conn_status, conn_json = conn.get_data(peripheral.picked_peripheral)
            self._session_token = conn.session_token

            if conn_status == "OK":
                peripheral.support_code = conn_json["results"]["DataContainer"][0][
                    "ContDetails"
                ][0]["LValue"]
                peripheral.status_code = conn_json["results"]["DataContainer"][0][
                    "ContDetails"
                ][1]["LValue"]
            else:
                raise RuntimeError(
                    f"An error occurred when refreshing status: {conn_status}"
                )

    def set_status(self, status_name: str, status_value: int, device_name: str) -> bool:
        """Set status to a peripheral.

        Parameters
        ----------
        status_name : str
            Status name, which has to be in idx dict. E.g. JciHitachiAC.idx
        status_value : int
            Status value.
        device_name : str
            Device name.

        Returns
        -------
        bool
            Return True if the command has been successfully executed. Otherwise, return False.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._sync_peripherals_availablity()
        if not self._peripherals[device_name].available:
            return False

        commander = self._peripherals[device_name].commander

        conn = connection.CreateJob(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response,
        )
        conn2 = connection.GetJobDoneReport(
            self.email,
            self.password,
            session_token=self._session_token,
            print_response=self.print_response,
        )

        # The mqtt events occurring order:
        # job (occurs one time) ->
        # job_done_report (occurs many times until the job status is successful) ->
        # peripheral (occurs one time, if this step is timed out, the job command fails)
        self._mqtt.mqtt_events.job.clear()
        self._mqtt.mqtt_events.job_done_report.clear()
        self._mqtt.mqtt_events.peripheral.clear()

        conn_status, conn_json = conn.get_data(
            gateway_id=self._peripherals[device_name].gateway_id,
            device_id=self._device_id,
            task_id=self.task_id,
            job_info=commander.get_b64command(status_name, status_value),
        )
        self._session_token = conn.session_token

        if not self._mqtt.mqtt_events.job.wait(timeout=10.0):
            return False
        for _ in range(self.max_retries):
            if not self._mqtt.mqtt_events.job_done_report.wait(timeout=10.0):
                self._mqtt.mqtt_events.job_done_report.clear()
                continue
            conn_status, conn_json = conn2.get_data(device_id=self._device_id)
            if (
                conn_status == "OK"
                and len(conn_json["results"]) != 0
                and conn_json["results"][0]["JobStatus"] == 0
            ):
                if not self._mqtt.mqtt_events.peripheral.wait(timeout=10.0):
                    return False
                time.sleep(0.5)  # still needs to wait a moment.
                return True
        return False


class AWSThing:
    """AWS thing (device) information.

    Parameters
    ----------
    thing_json : dict
        Thing json of a specific device.
    """

    supported_device_type = {
        "1": "AC",
        "2": "DH",
        "3": "HE",
        # "4": "PM25_PANEL",
    }

    def __init__(self, thing_json: dict) -> None:
        self._json: dict = thing_json
        self._available: bool = True
        self._shadow: Optional[dict] = None
        self._status_code: Optional[JciHitachiAWSStatus] = None
        self._support_code: Optional[JciHitachiAWSStatusSupport] = None
        self._monthly_data: Optional[list[dict]] = None

    def __repr__(self) -> str:
        ret = (
            f"name: {self.name}\n"
            f"brand: {self.brand}\n"
            f"model: {self.model}\n"
            f"type: {self.type}\n"
            f"firmware_code: {self.firmware_code}\n"
            f"firmawre_version: {self.firmware_version}\n"
            f"available: {self.available}\n"
            f"status_code: {self.status_code}\n"
            f"support_code: {self.support_code}\n"
            f"shadow: {self.shadow}\n"
            f"gateway_mac_address: {self.gateway_mac_address}"
        )
        return ret

    @classmethod
    def from_device_names(
        cls, things_json: dict, device_names: Optional[Union[list[str], str]]
    ) -> dict[str, object]:
        """Use device names to pick things_json accordingly.

        Parameters
        ----------
        things_json : dict
            things_json retrieved from aws_connection.GetAllDevice.
        device_names : list of str or str or None
            Device name. If None is given, all available devices will be included, by default None.

        Returns
        -------
        dict
            A dict of AWSThing instances with device name key.
        """

        things = {}

        if isinstance(device_names, str):
            device_names = [device_names]

        for thing in things_json["results"]["Things"]:
            device_name = thing["CustomDeviceName"]
            device_type = thing["DeviceType"]

            if device_names is None or (device_names and device_name in device_names):
                if device_type in cls.supported_device_type:
                    things[device_name] = cls(thing)

        assert device_names is None or len(device_names) == len(
            things
        ), "Some of device_names are not available from the API."

        return things

    @property
    def available(self) -> bool:
        """Whether the device is available.

        Returns
        -------
        bool
            Return True if the device is available.
        """

        return self._available

    @available.setter
    def available(self, x: bool) -> None:
        self._available = x

    @property
    def brand(self) -> str:
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """

        return getattr(self._support_code, "Brand")

    @property
    def firmware_version(self) -> str:
        """Firmware version.

        Returns
        -------
        str
            Device firmware version.
        """

        return getattr(self._support_code, "FirmwareVersion")

    @property
    def firmware_code(self) -> str:
        """Firmware code.

        Returns
        -------
        str
            Device firmware code.
        """

        return getattr(self._support_code, "FirmwareCode")

    @property
    def gateway_mac_address(self) -> str:
        """Gateway mac address.

        Returns
        -------
        str
            Gateway mac address.
        """

        return self._json["ThingName"].split("_")[-1]

    @property
    def model(self) -> str:
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        return getattr(self._support_code, "Model")

    @property
    def name(self) -> str:
        """Device name.

        Returns
        -------
        str
            Device name.
        """

        return self._json["CustomDeviceName"]

    @property
    def picked_thing(self) -> dict:
        """Picked thing.

        Returns
        -------
        dict
            Picked thing.
        """

        return self._json

    @property
    def shadow(self) -> Optional[dict]:
        """Thing's shadow reported by the API.

        Returns
        -------
        dict
            Shadow.
        """

        return self._shadow

    @shadow.setter
    def shadow(self, x: dict) -> None:
        self._shadow = x

    @property
    def status_code(self) -> Optional[JciHitachiAWSStatus]:
        """Thing's status code reported by the API.

        Returns
        -------
        JciHitachiAWSStatus
            Status code.
        """

        return self._status_code

    @status_code.setter
    def status_code(self, x: JciHitachiAWSStatus) -> None:
        self._status_code = x

    @property
    def support_code(self) -> Optional[JciHitachiAWSStatusSupport]:
        """Thing's support code reported by the API.

        Returns
        -------
        JciHitachiAWSStatusSupport
            Status code.
        """

        return self._support_code

    @support_code.setter
    def support_code(self, x: JciHitachiAWSStatusSupport) -> None:
        self._support_code = x

    @property
    def monthly_data(self) -> Optional[list[dict]]:
        """Thing's monthly data reported by the API.

        Returns
        -------
        Optional[list[dict]]
            Monthly data.
        """

        return self._monthly_data

    @monthly_data.setter
    def monthly_data(self, x: Optional[list[dict]]) -> None:
        self._monthly_data = x

    @property
    def thing_name(self) -> str:
        return self._json["ThingName"]

    @property
    def type(self) -> str:
        """Device type.

        Returns
        -------
        str
            Device type.
            If not supported, 'unknown' will be returned.
        """

        return self.supported_device_type.get(self._json["DeviceType"], "unknown")


class JciHitachiAWSAPI:
    """Jci-Hitachi API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    device_names : list of str or str or None, optional
        Device names. If None is given, all available devices will be included, by default None.
    max_retries : int, optional
        Maximum number of retries when setting status, by default 5.
    device_offline_timeout: float, optional
        For future use.
    print_response : bool, optional
        If set, all responses of httpx and MQTT will be printed, by default False.
    """

    def __init__(
        self,
        email: str,
        password: str,
        device_names: Optional[Union[list[str], str]] = None,
        max_retries: int = 5,
        device_offline_timeout: float = 10.0,
        print_response: bool = False,
    ) -> None:
        self.email: str = email
        self.password: str = password
        self.device_names: Optional[Union[list[str], str]] = device_names
        self.max_retries: int = max_retries
        self.print_response: bool = print_response

        self._mqtt: Optional[aws_connection.JciHitachiAWSMqttConnection] = None
        self._mqtt_timeout: float = device_offline_timeout
        self._shadow_names: Union[str, list] = ["info"]
        self._device_id: int = random.randint(1000, 6999)
        self._things: dict[str, AWSThing] = {}
        self._aws_tokens: Optional[aws_connection.AWSTokens] = None
        self._aws_identity: Optional[aws_connection.AWSIdentity] = None
        self._task_id: int = 0

    @property
    def things(self) -> dict[str, AWSThing]:
        """Picked things.

        Returns
        -------
        dict of AWSThing
            A dict of AWSThing instances.
        """

        return self._things

    @property
    def task_id(self) -> int:
        """Task ID.

        Returns
        -------
        int
            Serial number counted from 0, with maximum 999.
        """

        self._task_id += 1
        if self._task_id >= 1000:
            self._task_id = 1

        return self._task_id

    def _check_before_publish(self) -> None:
        # Reauthenticate 5 mins before AWSTokens expiration.
        current_time = time.time()
        if self._aws_tokens.expiration - current_time <= 300:
            self.reauth()

        if self._mqtt.mqtt_events.mqtt_error_event.is_set():
            self._mqtt.mqtt_events.mqtt_error_event.clear()
            self.reauth()

    def _get_valid_things(
        self, device_name: Optional[str] = None
    ) -> tuple[str, AWSThing]:
        for name, thing in self._things.items():
            if (device_name and name != device_name) or thing.type == "unknown":
                continue
            yield name, thing

    def _delay(self) -> None:
        time.sleep(0.2)

    def login(self) -> None:
        """Login API.

        Raises
        ------
        RuntimeError
            If a login error occurs, RuntimeError will be raised.
        """

        conn = aws_connection.GetUser(
            email=self.email,
            password=self.password,
            print_response=self.print_response,
        )
        self._aws_tokens = conn.aws_tokens
        conn_status, self._aws_identity = conn.get_data()

        conn = aws_connection.GetAllDevice(
            self._aws_tokens, print_response=self.print_response
        )
        conn_status, conn_json = conn.get_data()

        if conn_status == "OK":
            # things
            self._things = AWSThing.from_device_names(conn_json, self.device_names)
            self.device_names = list(self._things.keys())
            thing_names = [value.thing_name for value in self._things.values()]

            # mqtt
            def get_credential_callable():
                self._check_before_publish()
                conn = aws_connection.GetCredentials(
                    email=self.email,
                    password=self.password,
                    aws_tokens=self._aws_tokens,
                    print_response=self.print_response,
                )
                conn_status, aws_credentials = conn.get_data(self._aws_identity)
                if conn_status != "OK":
                    aws_connection._LOGGER.error(
                        f"An error occurred when acquiring a new AwsCredentials: {conn_status}"
                    )
                return aws_credentials

            self._mqtt = aws_connection.JciHitachiAWSMqttConnection(
                get_credential_callable, print_response=self.print_response
            )
            self._mqtt.configure(self._aws_identity.identity_id)

            if not self._mqtt.connect(
                self._aws_identity.host_identity_id, self._shadow_names, thing_names
            ):
                raise RuntimeError(
                    "An error occurred when connecting to MQTT endpoint."
                )

            # status
            self.refresh_status(refresh_support_code=True, refresh_shadow=True)
        else:
            raise RuntimeError(
                f"An error occurred when retrieving devices info: {conn_status}"
            )

    def logout(self) -> None:
        """Logout API."""

        self._mqtt.disconnect()

    def reauth(self) -> None:
        """Reauthenticate with AWS Cognito Service."""

        conn = aws_connection.JciHitachiAWSCognitoConnection(
            email=self.email,
            password=self.password,
            aws_tokens=self._aws_tokens,
            print_response=self.print_response,
        )
        conn_status, self._aws_tokens = conn.login(use_refresh_token=False)
        if conn_status != "OK":
            raise RuntimeError(
                f"An error occurred when reauthenticating with AWS Cognito Service: {conn_status}"
            )

    def change_password(self, new_password: str) -> None:
        """Change password.

            Warning:
            Use this function carefully, be sure you specify a strong enough password;
            otherwise, your password might be accepted by the Hitachi account management but not be accepted by AWS Cognito or vice versa,
            which will result in a login failure in the APP.

        Parameters
        ----------
        new_password : str
            New password.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        # We have to change the password in both AWS Cognito and Hitachi owned account management simultaneously.
        conn = aws_connection.ChangePassword(
            self.email,
            self.password,
            aws_tokens=self._aws_tokens,
            print_response=self.print_response,
        )
        aws_conn_status, _ = conn.get_data(new_password)
        if aws_conn_status != "OK":
            raise RuntimeError(
                f"An error occurred when changing AWS Cognito password: {aws_conn_status}"
            )

        conn = connection.UpdateUserCredential(
            self.email, self.password, print_response=self.print_response
        )
        hitachi_conn_status, _ = conn.get_data(new_password)
        if hitachi_conn_status != "OK":
            raise RuntimeError(
                f"An error occurred when changing Hitachi password: {hitachi_conn_status}"
            )

    def refresh_monthly_data(self, months: int, device_name: str) -> None:
        """Refresh available monthly data (power consumption) from the API.

        Parameters
        ----------
        months : int
            Number of months to get.
        device_name : str
            Device name.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        thing = self._things[device_name]
        current_timestamp_millis = time.time() * 1000

        conn = aws_connection.GetAvailableAggregationMonthlyData(
            self._aws_tokens, print_response=self.print_response
        )

        conn_status, response = conn.get_data(
            thing.thing_name,
            int(
                current_timestamp_millis - months * 2678400000
            ),  # 2678400000 ms == 31 days
            int(current_timestamp_millis),
        )
        if conn_status != "OK":
            raise RuntimeError(
                f"An error occurred when getting monthly data: {conn_status}"
            )

        thing.monthly_data = sorted(
            response["results"]["Data"], key=lambda x: x["Timestamp"]
        )

    def refresh_status(
        self,
        device_name: Optional[str] = None,
        refresh_support_code: bool = False,
        refresh_shadow: bool = False,
    ) -> None:
        """Refresh device status from the API.

        Parameters
        ----------
        device_name : str, optional
            Refreshing a device's status by its name.
            If None is given, all devices' status will be refreshed,
            by default None.
        refresh_support_code : bool, optional
            Whether or not to refresh support code, by default False.
        refresh_shadow : bool, optional
            Whether or not to refresh AWS IoT Shadow, by default False.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        # queue tasks
        for name, thing in self._get_valid_things(device_name):
            self._check_before_publish()

            if refresh_support_code:
                self._mqtt.publish(
                    self._aws_identity.host_identity_id,
                    thing.thing_name,
                    "support",
                    self._mqtt_timeout,
                )
            if refresh_shadow:
                self._mqtt.publish_shadow(thing.thing_name, "get", shadow_name="info")

            self._mqtt.publish(
                self._aws_identity.host_identity_id,
                thing.thing_name,
                "status",
                self._mqtt_timeout,
            )

        # execute
        support_results, shadow_results, status_results, _ = self._mqtt.execute()

        # gather results
        for name, thing in self._get_valid_things(device_name):
            if refresh_support_code:
                if thing.thing_name in support_results:
                    if thing.thing_name not in self._mqtt.mqtt_events.device_support:
                        raise RuntimeError(
                            f"An event occurred but wasn't accompanied with data when refreshing {name} support code."
                        )
                    thing.support_code = self._mqtt.mqtt_events.device_support[
                        thing.thing_name
                    ]
                else:
                    raise RuntimeError(
                        f"Timed out refreshing {name} support code. Please ensure the device is online and avoid opening the official app."
                    )
            if refresh_shadow:
                if thing.thing_name in shadow_results:
                    if thing.thing_name not in self._mqtt.mqtt_events.device_shadow:
                        raise RuntimeError(
                            f"An event occurred but wasn't accompanied with data when refreshing {name} shadow."
                        )
                    thing.shadow = self._mqtt.mqtt_events.device_shadow[
                        thing.thing_name
                    ]
                else:
                    raise RuntimeError(
                        f"Timed out refreshing {name} shadow. Please ensure the device is online and avoid opening the official app."
                    )

            if thing.thing_name in status_results:
                if thing.thing_name not in self._mqtt.mqtt_events.device_status:
                    raise RuntimeError(
                        f"An event occurred but wasn't accompanied with data when refreshing {name} status code."
                    )
                thing.status_code = self._mqtt.mqtt_events.device_status[
                    thing.thing_name
                ]
            else:
                raise RuntimeError(
                    f"Timed out refreshing {name} status code. Please ensure the device is online and avoid opening the official app."
                )

    def get_status(
        self, device_name: Optional[str] = None, legacy: bool = False
    ) -> dict[str, JciHitachiAWSStatus]:
        """Get device status after refreshing status.

        Parameters
        ----------
        device_name : str, optional
            Getting a device's status by its name.
            If None is given, all devices' status will be returned,
            by default None.
        legacy : bool, optional
            Whether or not to return status with legacy status name, by default False.

        Returns
        -------
        dict of JciHitachiAWSStatus.
            A dict of JciHitachiAWSStatus instances.
        """

        statuses = {}
        for name, thing in self._get_valid_things(device_name):
            if legacy:
                statuses[name] = thing.status_code.legacy_status
            else:
                statuses[name] = thing.status_code

            # inject temp and humidity limitations from the support code
            if thing.type == "AC":
                statuses[name]._status["max_temp"] = thing.support_code.max_temp
                statuses[name]._status["min_temp"] = thing.support_code.min_temp
            elif thing.type == "DH":
                statuses[name]._status["max_humidity"] = thing.support_code.max_humidity
                statuses[name]._status["min_humidity"] = thing.support_code.min_humidity

        return statuses

    def set_status(
        self,
        status_name: str,
        device_name: str,
        status_value: int = None,
        status_str_value: str = None,
    ) -> bool:
        """Set status to a thing. Either status_value or status_str_value must be specified.

        Parameters
        ----------
        status_name : str
            Status name.
        device_name : str
            Device name.
        status_value : int, optional
            Status value, by default None.
        status_str_value : str, optional
            Status string value, by default None.

        Returns
        -------
        bool
            Return True if the command has been successfully executed. Otherwise, return False.

        Raises
        ------
        RuntimeError
            If an error occurs, RuntimeError will be raised.
        """

        self._check_before_publish()

        thing = self._things[device_name]

        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id(
            device_type=thing.type,
            status_name=status_name,
            status_value=status_value,
            status_str_value=status_str_value,
            support_code=thing.support_code,
        )

        if not is_valid:
            return False

        shadow_publish_mapping = {
            "CleanFilterNotification": "filter",
            "CleanNotification": "filter",
            "CleanSecondaryFilterNotification": "filter",
            "FrontFilterNotification": "filter",
            "Pm25FilterNotification": "filter",
            "enableQAMode": "qa",
        }

        if False:  # status_name in shadow_publish_mapping: # TODO: replace False cond after shadow function is completed.
            shadow_publish_schema = {}
            if (
                shadow_publish_mapping[status_name] == "filter"
                or shadow_publish_mapping[status_name] == "qa"
            ):
                shadow_publish_schema.update({status_name: bool(status_value)})
                # if thing.type == "AC": # there is an additional parameter for AC
                #    shadow_publish_schema.update("FilterElapsedHour", 0 if status_value == 0 else status_value)
            self._mqtt.publish_shadow(
                thing.thing_name,
                "update",
                {"state": {"reported": {**shadow_publish_schema}}},
                shadow_name="info",
            )
            if self._mqtt.mqtt_events.device_control_event.wait(
                timeout=self._mqtt_timeout
            ):
                device_control = self._mqtt.mqtt_events.device_control.get(
                    thing.thing_name
                )
                if device_control["state"]["reported"][status_name] == bool(
                    status_value
                ):
                    self._mqtt.mqtt_events.device_control_event.clear()
                    self._delay()
                    return True
            return False

        self._mqtt.publish(
            self._aws_identity.host_identity_id,
            thing.thing_name,
            "control",
            self._mqtt_timeout,
            {
                status_name: status_value,
                "TaskID": self.task_id,
                "Timestamp": int(time.time()),
            },
        )

        _, _, _, control_results = self._mqtt.execute(control=True)

        if thing.thing_name in control_results:
            device_control = self._mqtt.mqtt_events.device_control.get(thing.thing_name)
            if device_control.get(status_name) == status_value:
                thing.status_code.set_new_status(status_name, status_value)
                return True
        return False



import concurrent
import datetime
import time
import threading
from unittest.mock import MagicMock, patch

import awscrt
import awsiot
import pytest

from JciHitachi.aws_connection import (
    AWS_COGNITO_ENDPOINT,
    AWS_COGNITO_IDP_ENDPOINT,
    AWS_IOT_ENDPOINT,
    AWSIdentity,
    AWSTokens,
    ChangePassword,
    GetAllDevice,
    GetAllGroup,
    GetAllRegion,
    GetAvailableAggregationMonthlyData,
    GetCredentials,
    GetHistoryEventByUser,
    GetUser,
    JciHitachiAWSCognitoConnection,
    JciHitachiAWSMqttConnection,
    ListSubUser,
)
from JciHitachi.model import JciHitachiAWSStatus, JciHitachiAWSStatusSupport

from . import MOCK_GATEWAY_MAC


@pytest.fixture()
def fixture_aws_tokens():
    return AWSTokens("access_token", "id_token", "refresh_token", time.time() + 3600)


@pytest.fixture()
def fixture_aws_mock_mqtt_connection():
    def get_credentials_callable():
        return awscrt.auth.AwsCredentials(
            access_key_id="",
            secret_access_key="",
            session_token="",
            expiration=datetime.datetime.fromtimestamp(time.time()),
        )

    mqtt_connection = JciHitachiAWSMqttConnection(
        get_credentials_callable, print_response=True
    )
    return mqtt_connection


class TestJciHitachiAWSMqttConnection:
    def test_on_publish_callback(self, fixture_aws_mock_mqtt_connection):
        mqtt = fixture_aws_mock_mqtt_connection
        thing_name = (
            f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}"
        )

        status_topic = f"{thing_name.split('_')[0]}/{thing_name}/status/response"
        mqtt._mqtt_events.device_status_event[thing_name] = threading.Event()
        mqtt._on_publish(status_topic, b'{"DeviceType": 1}', None, None, None)
        assert isinstance(
            mqtt._mqtt_events.device_status[thing_name], JciHitachiAWSStatus
        )
        assert mqtt._mqtt_events.device_status_event[thing_name].is_set()

        registration_topic = (
            f"{thing_name.split('_')[0]}/{thing_name}/registration/response"
        )
        mqtt._mqtt_events.device_support_event[thing_name] = threading.Event()
        mqtt._on_publish(registration_topic, b'{"DeviceType": 1}', None, None, None)
        assert isinstance(
            mqtt._mqtt_events.device_support[thing_name], JciHitachiAWSStatusSupport
        )
        assert mqtt._mqtt_events.device_support_event[thing_name].is_set()

        control_topic = f"{thing_name.split('_')[0]}/{thing_name}/control/response"
        mqtt._mqtt_events.device_control_event[thing_name] = threading.Event()
        mqtt._on_publish(control_topic, b'{"DeviceType": 1}', None, None, None)
        assert isinstance(mqtt._mqtt_events.device_control[thing_name], dict)
        assert mqtt._mqtt_events.device_control_event[thing_name].is_set()

        mqtt._mqtt_events.mqtt_error_event.clear()
        mqtt._on_publish("", b"", None, None, None)
        assert mqtt._mqtt_events.mqtt_error == "JSONDecodeError"
        assert mqtt._mqtt_events.mqtt_error_event.is_set()

    def test_on_get_named_shadow_accepted_callback(
        self, fixture_aws_mock_mqtt_connection
    ):
        mqtt = fixture_aws_mock_mqtt_connection
        thing_name = (
            f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}"
        )
        client_token = thing_name.split("_")[1]
        mqtt._client_tokens = {client_token: thing_name}

        response = MagicMock()
        response.client_token = client_token
        response.state.reported = {"Something": 0}
        mqtt._mqtt_events.device_shadow_event[thing_name] = threading.Event()
        mqtt._on_get_named_shadow_accepted(response)
        assert len(mqtt._client_tokens) == 0
        assert isinstance(mqtt._mqtt_events.device_shadow[thing_name], dict)
        assert mqtt._mqtt_events.device_shadow_event[thing_name].is_set()

        response.client_token = ""
        mqtt._on_get_named_shadow_accepted(response)

    def test_on_update_named_shadow_accepted_callback(
        self, fixture_aws_mock_mqtt_connection
    ):
        mqtt = fixture_aws_mock_mqtt_connection
        thing_name = (
            f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}"
        )
        client_token = thing_name.split("_")[1]
        mqtt._client_tokens = {client_token: thing_name}

        response = MagicMock()
        response.client_token = client_token
        response.state.reported = {"Something": 0}
        mqtt._mqtt_events.device_control_event[thing_name] = threading.Event()
        mqtt._on_update_named_shadow_accepted(response)
        assert len(mqtt._client_tokens) == 0
        assert isinstance(mqtt._mqtt_events.device_control[thing_name], dict)
        assert mqtt._mqtt_events.device_control_event[thing_name].is_set()

        response.client_token = ""
        mqtt._on_update_named_shadow_accepted(response)

    def test_configure(self, fixture_aws_mock_mqtt_connection):
        mqtt = fixture_aws_mock_mqtt_connection

        assert mqtt._mqttc is None
        assert mqtt._shadow_mqttc is None

        mqtt.configure(identity_id="identity_id")

        assert isinstance(mqtt._mqttc, awscrt.mqtt.Connection)
        assert isinstance(mqtt._shadow_mqttc, awsiot.iotshadow.IotShadowClient)

    def test_connect(self, fixture_aws_mock_mqtt_connection):
        mqtt = fixture_aws_mock_mqtt_connection

        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            connect_future = concurrent.futures.Future()
            connect_future.set_result(None)
            mock_mqttc.connect.return_value = connect_future
            subscribe_future = concurrent.futures.Future()
            subscribe_future.set_result(None)
            mock_mqttc.subscribe.return_value = (subscribe_future, None)
            assert mqtt.connect("")

        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            connect_future = concurrent.futures.Future()
            connect_future.set_exception(ValueError())
            mock_mqttc.connect.return_value = connect_future
            mqtt._mqtt_events.mqtt_error = ""
            mqtt._mqtt_events.mqtt_error_event.clear()
            assert not mqtt.connect("")
            assert mqtt._mqtt_events.mqtt_error == "ValueError"
            assert mqtt._mqtt_events.mqtt_error_event.is_set()

        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            subscribe_future = concurrent.futures.Future()
            subscribe_future.set_exception(RuntimeError())
            mock_mqttc.subscribe.return_value = (subscribe_future, None)
            mqtt._mqtt_events.mqtt_error = ""
            mqtt._mqtt_events.mqtt_error_event.clear()
            assert not mqtt.connect("")
            assert mqtt._mqtt_events.mqtt_error == "RuntimeError"
            assert mqtt._mqtt_events.mqtt_error_event.is_set()

    def test_publish(self, fixture_aws_mock_mqtt_connection):
        mqtt = fixture_aws_mock_mqtt_connection
        thing_name = (
            f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}"
        )

        assert thing_name not in mqtt._mqtt_events.device_support_event
        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            publish_future = concurrent.futures.Future()
            publish_future.set_result(None)
            mock_mqttc.publish.return_value = (publish_future, None)
            mqtt.publish("", thing_name, "support")
            assert thing_name in mqtt._mqtt_events.device_support_event
            assert len(mqtt._execution_pools.support_execution_pool) == 1

        # test clearing event
        mqtt._mqtt_events.device_support_event[thing_name] = threading.Event()
        mqtt._mqtt_events.device_support_event[thing_name].set()
        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            publish_future = concurrent.futures.Future()
            publish_future.set_result(None)
            mock_mqttc.publish.return_value = (publish_future, None)
            mqtt.publish("", thing_name, "support")
            assert thing_name in mqtt._mqtt_events.device_support_event
            assert not mqtt._mqtt_events.device_support_event[thing_name].is_set()

        # test invalid publish_type
        with patch.object(mqtt, "_mqttc") as mock_mqttc:
            publish_future = concurrent.futures.Future()
            publish_future.set_exception(ValueError())
            mock_mqttc.publish.return_value = (publish_future, None)
            with pytest.raises(ValueError, match="Invalid publish_type: others"):
                mqtt.publish("", thing_name, "others")

        # TODO: test timeout

    @pytest.mark.parametrize("raise_exception", [False, True])
    def test_publish_shadow(self, fixture_aws_mock_mqtt_connection, raise_exception):
        mqtt = fixture_aws_mock_mqtt_connection
        thing_name = (
            f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}"
        )

        with patch.object(mqtt, "_shadow_mqttc") as mock_shadow_mqttc:

            def publish_shadow_func(request, qos):
                assert isinstance(request, awsiot.ModeledClass)
                assert isinstance(qos, int)
                publish_future = concurrent.futures.Future()
                if raise_exception:
                    publish_future.set_exception(RuntimeError())
                else:
                    publish_future.set_result(None)
                return publish_future

            shadow_types_to_test = [
                (mock_shadow_mqttc.publish_get_shadow, "get", None),
                (mock_shadow_mqttc.publish_update_shadow, "update", None),
                (mock_shadow_mqttc.publish_get_named_shadow, "get", "info"),
                (mock_shadow_mqttc.publish_update_named_shadow, "update", "info"),
            ]

            for shadow_func, command_name, shadow_name in shadow_types_to_test:
                shadow_func.side_effect = publish_shadow_func

                mqtt._mqtt_events.mqtt_error = ""
                mqtt._mqtt_events.mqtt_error_event.clear()
                if raise_exception:
                    pass
                    # mqtt.publish_shadow(thing_name, command_name, shadow_name=shadow_name)
                    # assert mqtt._mqtt_events.mqtt_error == "RuntimeError"
                    # assert mqtt._mqtt_events.mqtt_error_event.is_set()
                else:
                    mqtt.publish_shadow(
                        thing_name, command_name, shadow_name=shadow_name
                    )
                    assert not mqtt._mqtt_events.mqtt_error_event.is_set()

        # Test invalid command name.
        with pytest.raises(
            ValueError, match="command_name must be one of `get` or `update`."
        ):
            mqtt.publish_shadow(thing_name, "delete")

    def test_execute(self, fixture_aws_mock_mqtt_connection):
        mqtt = fixture_aws_mock_mqtt_connection
        # The executed functions are no-op.
        mqtt._execution_pools.support_execution_pool.append(
            mqtt._wrap_async("support_identifier", lambda: None)
        )
        mqtt._execution_pools.shadow_execution_pool.append(
            mqtt._wrap_async("shadow_identifier", lambda: None)
        )
        mqtt._execution_pools.status_execution_pool.append(
            mqtt._wrap_async("status_identifier", lambda: None)
        )
        mqtt._execution_pools.control_execution_pool.append(
            mqtt._wrap_async("control_identifier", lambda: None)
        )

        results = mqtt.execute()
        assert results == (
            ["support_identifier"],
            ["shadow_identifier"],
            ["status_identifier"],
            None,
        )
        assert len(mqtt._execution_pools.support_execution_pool) == 0
        assert len(mqtt._execution_pools.shadow_execution_pool) == 0
        assert len(mqtt._execution_pools.status_execution_pool) == 0

        results = mqtt.execute(control=True)
        assert results == (None, None, None, ["control_identifier"])
        assert len(mqtt._execution_pools.control_execution_pool) == 0


class TestJciHitachiAWSCognitoConnection:
    # (class name, get data args, header_target, response json, response type)
    classes_to_test = [
        (
            JciHitachiAWSCognitoConnection,
            None,
            "AWSCognitoIdentityProviderService.InitiateAuth",
            {
                "AuthenticationResult": {
                    "AccessToken": "acc_token",
                    "IdToken": "IdToken",
                    "RefreshToken": "RefreshToken",
                    "ExpiresIn": 3600,
                }
            },
            AWSTokens,
        ),
        (
            ChangePassword,
            {"new_password": "x"},
            "AWSCognitoIdentityProviderService.ChangePassword",
            {},
            type(None),
        ),
        (
            GetUser,
            {},
            "AWSCognitoIdentityProviderService.GetUser",
            {
                "Username": "username",
                "UserAttributes": [
                    {
                        "Name": "custom:cognito_identity_id",
                        "Value": "identity_id",
                    },
                    {
                        "Name": "custom:host_identity_id",
                        "Value": "host_identity_id",
                    },
                ],
            },
            AWSIdentity,
        ),
        (
            GetCredentials,
            {"aws_identity": AWSIdentity("", "", "", {})},
            "AWSCognitoIdentityService.GetCredentialsForIdentity",
            {
                "Credentials": {
                    "AccessKeyId": "acc",
                    "SecretKey": "sec",
                    "SessionToken": "ses",
                    "Expiration": time.time(),
                }
            },
            awscrt.auth.AwsCredentials,
        ),
    ]
    error_response_json = {
        # For invalid requests
        "__type": "NotAuthorizedException",
        "message": "Incorrect username or password.",
    }
    login_response_json = {
        # For login
        "AuthenticationResult": {
            "AccessToken": "abc",
            "IdToken": "def",
            "RefreshToken": "ghi",
            "ExpiresIn": 3600,
        },
    }

    @pytest.mark.parametrize("test_class", classes_to_test)
    def test_get_data_or_login(self, test_class):
        (
            test_class,
            get_data_args,
            header_target,
            response_json,
            response_type,
        ) = test_class
        c = test_class(
            "abc@abc.com",
            "password",
            aws_tokens=AWSTokens("", "", "", time.time()),
            print_response=True,
        )
        with patch("httpx.post") as mock_post:
            response_code = [
                # (code, message)
                (200, "OK"),
                (
                    400,
                    f"{self.error_response_json['__type']} {self.error_response_json['message']}",
                ),
            ]

            for http_status_code, msg in response_code:

                def mock_post_func(
                    endpoint, headers=None, json=None, proxy=None, **kwargs
                ):
                    if isinstance(c, GetCredentials):
                        assert endpoint == f"https://{AWS_COGNITO_ENDPOINT}/"
                    else:
                        assert endpoint == f"https://{AWS_COGNITO_IDP_ENDPOINT}/"
                    assert headers == {
                        "X-Amz-Target": header_target,
                        "User-Agent": "Dalvik/2.1.0",
                        "content-type": "application/x-amz-json-1.1",
                        "Accept": "application/json",
                    }

                    response = MagicMock()
                    response.headers = ""
                    response.status_code = http_status_code
                    if http_status_code == 200:
                        response.json.return_value = response_json
                    else:
                        response.json.return_value = self.error_response_json
                    return response

                mock_post.side_effect = mock_post_func
                if c.__class__.__name__ == "JciHitachiAWSCognitoConnection":
                    response_msg, response = c.login()
                    with pytest.raises(NotImplementedError):
                        c.get_data()
                else:
                    response_msg, response = c.get_data(**get_data_args)

                if http_status_code == 200:
                    assert response_msg == msg
                    assert isinstance(response, response_type)
                else:
                    assert response_msg == msg
                    assert response is None


class TestJciHitachiIoTConnection:
    # (class name, get data args)
    classes_to_test = [
        (GetAllDevice, {}),
        (GetAllGroup, {}),
        (GetAllRegion, {}),
        (
            GetAvailableAggregationMonthlyData,
            {"thing_name": "", "time_start": "", "time_end": ""},
        ),
        (GetHistoryEventByUser, {"time_start": "", "time_end": ""}),
        (ListSubUser, {}),
    ]
    no_need_access_token = ["GetAllDevice", "GetAllRegion"]

    @pytest.mark.parametrize("test_class", classes_to_test)
    def test_generate_headers(self, test_class, fixture_aws_tokens):
        test_class, get_data_args = test_class
        c = test_class(fixture_aws_tokens, print_response=True)

        headers = {
            "authorization": f"Bearer {fixture_aws_tokens.id_token}",
            "User-Agent": "Dalvik/2.1.0",
            "content-type": "application/json",
            "Accept": "application/json",
        }

        if test_class.__name__ not in self.no_need_access_token:
            headers["accesstoken"] = f"Bearer {fixture_aws_tokens.access_token}"
            assert c._generate_headers(True) == headers
        else:
            assert c._generate_headers(False) == headers

    @pytest.mark.parametrize("test_class", classes_to_test)
    def test_get_data(self, test_class, fixture_aws_tokens):
        test_class, get_data_args = test_class
        c = test_class(fixture_aws_tokens, print_response=True)
        with patch("httpx.post") as mock_post:
            response_code = [
                # (code, message)
                (0, "OK"),
                (6, "Invalid email or password"),
                (12, "Invalid session token"),
                (100, "Unknown error"),
            ]

            for code, msg in response_code:
                response_json = {"status": {"code": code}}

                for http_status_code in [200, 400]:

                    def mock_post_func(
                        endpoint, headers=None, json=None, proxy=None, **kwargs
                    ):
                        h = {
                            "authorization": f"Bearer {fixture_aws_tokens.id_token}",
                            "User-Agent": "Dalvik/2.1.0",
                            "content-type": "application/json",
                            "Accept": "application/json",
                        }
                        if test_class.__name__ not in self.no_need_access_token:
                            h["accesstoken"] = (
                                f"Bearer {fixture_aws_tokens.access_token}"
                            )
                        assert (
                            endpoint
                            == f"https://{AWS_IOT_ENDPOINT}/{c.__class__.__name__}"
                        )
                        assert headers == h

                        response = MagicMock()
                        response.headers = ""
                        response.status_code = http_status_code
                        response.json = MagicMock(return_value=response_json)
                        return response

                    mock_post.side_effect = mock_post_func

                    if http_status_code == 200:
                        assert c.get_data(**get_data_args) == (msg, response_json)
                    else:
                        assert c.get_data(**get_data_args) == (
                            f"HTTP exception {http_status_code}",
                            response_json,
                        )



import base64

from . import utility as util
from .model import JciHitachiAC, JciHitachiDH, JciHitachiHE


class JciHitachiCommand:  # pragma: no cover
    """Abstract class for sending job command.

    Parameters
    ----------
    gateway_mac_address : str
        Gateway mac address.
    """

    def __init__(self, gateway_mac_address):
        self.job_info_base = bytearray.fromhex(
            "d0d100003c6a9dffff03e0d4ffffffff \
             00000100000000000000002000010000 \
             000000000000000002000d278050f0d4 \
             469dafd3605a6ebbdb130d278052f0d4 \
             469dafd3605a6ebbdb13060006000000 \
             0000"
        )
        self.job_info_base[32:40] = bytearray.fromhex(hex(int(gateway_mac_address))[2:])

    def get_command(self, command, value):
        raise NotImplementedError

    def get_b64command(self, command, value):
        """A wrapper of get_command, generating base64 command.

        Parameters
        ----------
        command : str
            Status name.
        value : int
            Status value.

        Returns
        -------
        str
            Base64 command.
        """

        return base64.b64encode(self.get_command(command, value)).decode()


class JciHitachiCommandAC(JciHitachiCommand):  # pragma: no cover
    """Sending job command to air conditioner.

    Parameters
    ----------
    gateway_mac_address : str
        Gateway mac address.
    """

    def __init__(self, gateway_mac_address):
        super().__init__(gateway_mac_address)

    def get_command(self, command, value):
        """Get job command.

        Parameters
        ----------
        command : str
            Status name.
        value : int
            Status value.

        Returns
        -------
        bytearray
            Bytearray command.
        """

        job_info = self.job_info_base.copy()

        # Device type
        job_info[77] = 1

        # Command (eg. target_temp)
        job_info[78] = 128 + JciHitachiAC.idx[command]

        # Value (eg. 27)
        job_info[80] = value

        # Checksum
        # Original algorithm:
        # xor job_info 76~80
        # Since byte 76(0x06), 77(device type), and 79(0x00) are constants,
        # here is the simplified algorithm:
        # command ^ value ^ 0x07 (flip last 3 bits)
        job_info[81] = job_info[78] ^ job_info[80] ^ 0x07

        return job_info


class JciHitachiCommandDH(JciHitachiCommand):  # pragma: no cover
    """Sending job command to dehumidifier.

    Parameters
    ----------
    gateway_mac_address : str
        Gateway mac address.
    """

    def __init__(self, gateway_mac_address):
        super().__init__(gateway_mac_address)

    def get_command(self, command, value):
        """Get job command.

        Parameters
        ----------
        command : str
            Status name.
        value : int
            Status value.

        Returns
        -------
        bytearray
            Bytearray command.
        """

        job_info = self.job_info_base.copy()

        # Device type
        job_info[77] = 4

        # Command (eg. target_temp)
        job_info[78] = 128 + JciHitachiDH.idx[command]

        # Value (eg. 27)
        job_info[80] = value

        # Checksum
        # Original algorithm:
        # xor job_info 76~80
        # Since byte 76(0x06), 77(device type), and 79(0x00) are constants,
        # here is the simplified algorithm:
        # command ^ value ^ 0x02
        job_info[81] = job_info[78] ^ job_info[80] ^ 0x02

        return job_info


class JciHitachiCommandHE(JciHitachiCommand):  # pragma: no cover
    """Sending job command to heat exchanger.

    Parameters
    ----------
    gateway_mac_address : str
        Gateway mac address.
    """

    def __init__(self, gateway_mac_address):
        super().__init__(gateway_mac_address)

    def get_command(self, command, value):
        """Get job command.

        Parameters
        ----------
        command : str
            Status name.
        value : int
            Status value.

        Returns
        -------
        bytearray
            Bytearray command.
        """

        job_info = self.job_info_base.copy()

        # Device type
        job_info[77] = 14

        # Command (eg. target_temp)
        job_info[78] = 128 + JciHitachiHE.idx[command]

        # Value (eg. 27)
        job_info[80] = value

        # Checksum
        # Original algorithm:
        # xor job_info 76~80
        # Since byte 76(0x06), 77(device type), and 79(0x00) are constants,
        # here is the simplified algorithm:
        # command ^ value ^ 0x08
        job_info[81] = job_info[78] ^ job_info[80] ^ 0x08

        return job_info


class JciHitachiStatusInterpreter:  # pragma: no cover
    """Interpreting received status code.

    Parameters
    ----------
    code : str
        status code.
    """

    def __init__(self, code):
        self.base64_bytes = base64.standard_b64decode(code)

    def _decode_status_number(self):
        if 6 < self.base64_bytes[0] and (
            self.base64_bytes[1],
            self.base64_bytes[2],
        ) == (0, 8):
            return int((self.base64_bytes[0] - 4) / 3)
        else:
            return 0

    def _decode_support_number(self):
        if 9 < self.base64_bytes[0]:
            return int((self.base64_bytes[0] - 26) / 3)
        else:
            return 0

    def _decode_single_status(self, max_func_number, while_counter):
        stat_idx = while_counter * 3 + 3

        if stat_idx + 3 <= self.base64_bytes[0] - 1:
            status_bytes = bytearray(4)
            status_bytes[0] = (self.base64_bytes[stat_idx] & 0x80) != 0
            status_bytes[1] = self.base64_bytes[stat_idx] & 0xFFFF7FFF
            status_bytes[2:4] = self.base64_bytes[stat_idx + 1 : stat_idx + 3]

            output = int.from_bytes(status_bytes, byteorder="little")
            return output
        else:
            output = util.bin_concat(0xFF, max_func_number)
            output = (output << 16) & 0xFFFF0000 | max_func_number
        return output

    def _decode_single_support(self, max_func_number, while_counter, init):
        stat_idx = while_counter * 3 + init

        if stat_idx + 3 <= self.base64_bytes[0] - 1:
            status_bytes = bytearray(4)
            status_bytes[0] = (self.base64_bytes[stat_idx] & 0x80) != 0
            status_bytes[1] = self.base64_bytes[stat_idx] & 0xFFFF7FFF
            status_bytes[2:4] = self.base64_bytes[stat_idx + 1 : stat_idx + 3]

            output = int.from_bytes(status_bytes, byteorder="little")
        else:
            output = util.bin_concat(0xFF, max_func_number)
            output = (output << 16) & 0xFFFF0000 | max_func_number
        return output

    def _get_strs(self, start_idx, num_strs):
        strs = []
        idx = start_idx

        while len(strs) < num_strs:
            char = self.base64_bytes[idx]
            if char == 0:
                strs.append(self.base64_bytes[start_idx:idx].decode())
                idx += 1
                start_idx = idx
            else:
                idx += 1
        return idx, strs

    def decode_status(self):
        """Decode all status codes of a peripheral.

        Returns
        -------
        dict
            Decoded status.
        """

        table = {}
        num_idx = self._decode_status_number()
        for i in range(num_idx):
            ret = self._decode_single_status(num_idx, i)
            idx = util.cast_bytes(ret >> 8, 1)
            table[idx] = ret >> 0x18 + (ret >> 0x10 * 0x100)
        return table

    def decode_support(self):
        """Decode all support codes of a peripheral.

        Returns
        -------
        dict
            Decoded support.
        """

        init, (brand, model) = self._get_strs(8, 2)

        table = {"brand": brand, "model": model}

        num_idx = self._decode_support_number()

        for i in range(num_idx):
            ret = self._decode_single_support(num_idx, i, init)
            idx = util.cast_bytes(ret >> 8, 1)
            if idx >= 128:
                idx = idx - 128
            # save raw value, the extraction procedure is performed in model.
            table[idx] = ret

        return table



import pytest

from JciHitachi.model import (
    JciHitachiAC,
    JciHitachiACSupport,
    JciHitachiDH,
    JciHitachiDHSupport,
    JciHitachiHE,
    JciHitachiHESupport,
)
from JciHitachi.status import (
    JciHitachiCommandAC,
    JciHitachiCommandDH,
    JciHitachiCommandHE,
    JciHitachiStatusInterpreter,
)

from . import (
    MOCK_CODE_AC,
    MOCK_CODE_DH,
    MOCK_GATEWAY_MAC,
    MOCK_SUPPORT_CODE_AC,
    MOCK_SUPPORT_CODE_DH,
    MOCK_COMMAND_AC,
    MOCK_COMMAND_DH,
)


class TestACStatus:
    def test_model_brand(self):
        dev_support = JciHitachiStatusInterpreter(MOCK_SUPPORT_CODE_AC).decode_support()
        ac_support = JciHitachiACSupport(dev_support)

        assert ac_support.brand == "HITACHI"
        assert ac_support.model == "RAD-90NF"

    def test_supported_status(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_AC).decode_status()
        ac_status = JciHitachiAC(dev_status).status

        for status in ac_status.values():
            assert status != "unknown" and status != None

    def test_unsupported_status(self):
        dev_status = {}
        ac_status = JciHitachiAC(dev_status).status

        for status in ac_status.values():
            assert status == "unsupported" or status == -1

    def test_correctness(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_AC).decode_status()
        ac_status = JciHitachiAC(dev_status).status

        mock_status = {
            "power": "off",
            "mode": "cool",
            "air_speed": "silent",
            "target_temp": 26,
            "indoor_temp": 26,
            "sleep_timer": 0,
            "vertical_wind_swingable": "unsupported",
            "vertical_wind_direction": -1,
            "horizontal_wind_direction": "unsupported",
            "mold_prev": "disabled",
            "fast_op": "disabled",
            "energy_save": "disabled",
            "sound_prompt": "enabled",
            "outdoor_temp": 24,
            "power_kwh": 0.0,
            "freeze_clean": "off",
        }

        for key, value in ac_status.items():
            assert key in mock_status
            assert mock_status[key] == value

    def test_limit(self):
        dev_support = JciHitachiStatusInterpreter(MOCK_SUPPORT_CODE_AC).decode_support()
        ac_support = JciHitachiACSupport(dev_support)
        mock_status_raw = {
            "power": 0,
            "mode": 0,
            "air_speed": 1,
            "target_temp": 26,
            "indoor_temp": 26,
            "sleep_timer": 0,
            "vertical_wind_swingable": -1,
            "vertical_wind_direction": -1,
            "horizontal_wind_direction": -1,
            "mold_prev": 0,
            "fast_op": 0,
            "energy_save": 0,
            "sound_prompt": 0,
            "outdoor_temp": 24,
            "power_kwh": 0,
        }

        mock_status_raw_limited = {
            key: ac_support.limit(key, value) for key, value in mock_status_raw.items()
        }

        for key, raw_value in mock_status_raw.items():
            if raw_value == -1:
                raw_value = None
            if (
                key == "outdoor_temp"
            ):  # The supported `outdoor_temp` values seem not correct so the limited values are incorrect too. skipping this test.
                continue
            assert raw_value == mock_status_raw_limited[key]

    def test_command(self):
        commander = JciHitachiCommandAC(MOCK_GATEWAY_MAC)
        b64command = commander.get_command(MOCK_COMMAND_AC, 0)

        mock_command = bytearray.fromhex(
            "d0d100003c6a9dffff03e0d4ffffffff \
             00000100000000000000002000010000 \
             908D9859F3FD7f6c02000d278050f0d4 \
             469dafd3605a6ebbdb130d278052f0d4 \
             469dafd3605a6ebbdb13060006019E00 \
             0099"
        )
        assert b64command == mock_command


class TestDHStatus:
    def test_model_brand(self):
        dev_support = JciHitachiStatusInterpreter(MOCK_SUPPORT_CODE_DH).decode_support()
        dh_support = JciHitachiDHSupport(dev_support)

        assert dh_support.brand == "HITACHI"
        assert dh_support.model == "RD-360HH"

    def test_supported_status(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_DH).decode_status()
        dh_status = JciHitachiDH(dev_status).status

        for status in dh_status.values():
            assert status != "unknown" and status != None

    def test_unsupported_status(self):
        dev_status = {}
        ac_status = JciHitachiDH(dev_status).status
        for status in ac_status.values():
            assert status == "unsupported" or status == -1

    def test_correctness(self):
        dev_status = JciHitachiStatusInterpreter(MOCK_CODE_DH).decode_status()
        ac_status = JciHitachiDH(dev_status).status

        mock_status = {
            "power": "on",
            "mode": "custom",
            "target_humidity": 70,
            "indoor_humidity": 70,
            "wind_swingable": "on",
            "water_full_warning": "off",
            "clean_filter_notify": "disabled",
            "air_purify_level": "unsupported",  # not implemented
            "air_speed": "silent",
            "side_vent": "off",
            "sound_control": "button+waterfull",
            "error_code": 0,
            "mold_prev": "off",
            "power_kwh": 0.4,
            "air_quality_value": -1,  # not implemented
            "air_quality_level": "unsupported",  # not implemented
            "pm25_value": 26,
            "display_brightness": "bright",
            "odor_level": "middle",
            "air_cleaning_filter": "enabled",
        }

        for key, value in ac_status.items():
            assert key in mock_status
            assert mock_status[key] == value

    def test_command(self):
        commander = JciHitachiCommandDH(MOCK_GATEWAY_MAC)
        b64command = commander.get_command(MOCK_COMMAND_DH, 0)

        mock_command = bytearray.fromhex(
            "d0d100003c6a9dffff03e0d4ffffffff \
             00000100000000000000002000010000 \
             908D9859F3FD7f6c02000d278050f0d4 \
             469dafd3605a6ebbdb130d278052f0d4 \
             469dafd3605a6ebbdb13060006049000 \
             0092"
        )
        assert b64command == mock_command


class TestCommonStatus:
    def test_command_num_bytes(self):
        ac_commander = JciHitachiCommandAC(MOCK_GATEWAY_MAC)
        assert len(ac_commander.get_command("power", 0)) == 82
        dh_commander = JciHitachiCommandDH(MOCK_GATEWAY_MAC)
        assert len(dh_commander.get_command("power", 0)) == 82
        # he_commander = JciHitachiCommandHE(MOCK_GATEWAY_MAC)
        # assert len(he_commander.get_command("power", 0)) == 82

    # TODO: Add device availablity test
